# FILE: server.tcl
#
# A generic server that can be instantiated and customized.
#

em_class net_server {

  # Protocol to be used on connections established by this server
  # Need good check that doesn't depend on naming conventions
  const public variable protocol 

  # The alias by which this server identifies itself
  const public variable alias = {}

    # Log for logging messages
    public variable log {
        if {[catch {$log Append "Logging messages generated by $this."}]} {
            error "$log is not able to log messages."
        }
        if {[info exists logHandler]} {
            $logHandler delete
        }
        set logHandler [em_eventHandler new $log delete "$this delete"]
        if {$imadelog} {
            $oldlog delete
        }
        set oldlog $log
    }
    private variable oldlog = em_nop
    private variable logHandler
    private variable imadelog = 0

  # Boolean - does this server register with a services directory?
  const public variable register = 1

  # List of remote hostname patterns allowed to make connections.
  public variable allow = {} {
        $log Append "$this: New allow: $allow"
  }

  # List of remote hostname patterns forbidden to make connections.
  public variable deny = {} {
        $log Append "$this: New deny: $deny"
  }

  # Track whether the server instance owns, and must clean up
  # the log and protocol instances it uses.
  const private variable imadeprotocol = 0

  # The socket through which this server accepts connections
  private variable socket

  private variable port

  private variable myEventHandlers

  constructor {args} {
    eval $this configure $args
    if {![info exists log]} {
      # Provide a default log
      set log [em_log new -title "Server $this Log"]
      set imadelog 1
    }
    # Die when $log dies.
    lappend myEventHandlers [em_eventHandler new $log delete "$this delete"]
    if {![info exists protocol]} {
      # Provide a default protocol
      set protocol [net_protocol new -name "Default protocol"]
      set imadeprotocol 1
    }
    # Die when $protocol dies.
    lappend myEventHandlers [em_eventHandler new $protocol delete \
            "$this delete"]
  }

  # Start the server accepting connections port
  method Start { _port } { 
    if {[info exists socket]} {
      error "Server already started"
    }
$log Append "Before socket -server"
    set socket [socket -server [list $this VerifyHostAccess] \
	    -myaddr localhost $_port]  ;# Force localhost for now
$log Append "After socket -server"
    set port [lindex [fconfigure $socket -sockname] 2]
$log Append "After fconfigure -sockname"
    $log Append "Server $this started on port $port..."
    if {$register} {
        $this Register
    }
  }

  method VerifyHostAccess {accsock clientipaddr clientport} {
    set myinfo [fconfigure $accsock -sockname]
    set remoteinfo [fconfigure $accsock -peername]
    set myhostname [lindex $myinfo 1]
    set remotehostname [lindex $remoteinfo 1]
    if {![string match $myhostname $remotehostname]} {
      if {[string match $clientipaddr $remotehostname]} {
        $this RefuseConnection $accsock $remotehostname $clientport
        return
      }
      set allowconnection 0
      foreach pattern $allow {
        if {[string match $pattern $remotehostname]} {
          set allowconnection 1
          break
        }
      }
      foreach pattern $deny {
        if {[string match $pattern $remotehostname]} {
          set allowconnection 0
          break
        }
      }
      if {!$allowconnection} {
        $this RefuseConnection $accsock $remotehostname $clientport
        return
      }
    }
    # This causes all connections created by this server to share its log.
    # When the server created the log, that means when it dies, it will
    # kill its log which will kill all the connections -- against the
    # design which lets connections persist after their server dies.
    net_connection new -socket $accsock -log $log -protocol $protocol
  }

  private method RefuseConnection {accsock rhost cport} {
    puts $accsock "tell Connection refused"
    close $accsock
    $log Append "Server $this refused connection from $rhost:$cport"
  }

  # Returns true if server is accepting connections, false otherwise.
  method IsListening {} {
    return [info exists socket]
  }

  # Return port we're listening on
  method Port {} {
      if {[info exists port]} {
          return $port
      }
      error "server is not listening"
  }

  # Stop accepting connections
  method Stop {} {
    if {![info exists socket]} {
      error "Server not started"
    }
    $log Append "Stopping server $this ..."
    if {$register} {
        set socketinfo [fconfigure $socket -sockname]
        set port [lindex $socketinfo 2]
        $this Deregister
    }
    unset port
    close $socket
    unset socket
  }

    private variable account
    private variable qhandler
    private variable thandler

    private method Register {} {
        if {[catch {net_account new -hostname localhost -log $log} msg]} {
            $this RegisterFail $msg
            return
        }
        set account $msg
        lappend myEventHandlers [em_eventHandler new $account delete \
                [list $this RegisterFail "account died"]]
        if {[$account Ready]} {
            $this AccountReady
        } else {
            lappend myEventHandlers [em_eventHandler new $account Ready \
                    [list $this AccountReady]]
        }
    }

    method AccountReady {} {
        set qid [$account Send register [pid] $alias $port]
        set qhandler [em_eventHandler new $account Reply$qid \
                [list $this RegisterReply]]
        set thandler [em_eventHandler new $account Timeout$qid \
                [list $this RegisterFail "timeout on registration reply"]]
    }

    method RegisterReply {} {
        $qhandler delete
        unset qhandler
        $thandler delete
        unset thandler
        set reply [$account Get]
        switch -- [lindex $reply 0] {
            0 {
                $log Append "Server $this registered successfully"
            }
            default {
                $this RegisterFail "registration error"
            }
        }
    }

    method RegisterFail {msg} {
        set accountname [net_account DefaultAccountName]
        $log Append "Unable to register server with localhost:$accountname:
	$msg"
        $this delete
    }

    private method Deregister {} {
        if {![info exists account] || ![$account Ready]} {
            return
        }
        set qid [$account Send deregister [pid] $port]
        set qhandler [em_eventHandler new $account Reply$qid \
                [list $this DeregisterReply]]
        set thandler [em_eventHandler new $account Timeout$qid \
                [list $this DeregisterFail "timeout on deregistration reply"]]
    }

    method DeregisterReply {} {
        $qhandler delete
        unset qhandler
        $thandler delete
        unset thandler
        set reply [$account Get]
        switch -- [lindex $reply 0] {
            0 {
                $log Append "Server $this deregistered successfully"
            }
            default {
                $this RegisterFail "deregistration error"
            }
        }
    }

    method DeregisterFail {msg} {
        set accountname [net_account DefaultAccountName]
        $log Append "Unable to deregister server with localhost:$accountname:
	$msg"
        # Ought to send up fireworks; this is bad.
        $this delete
    }

  destructor {
    if {[info exists socket]} {
        # Stop method launches Deregister in background.  Should we wait
        # to confirm it before dying?   Proper usage is to Stop server,
        # then delete it.  More thought needed here...
        $this Stop
    }
    if {[info exists log]} {
        $log Append "Deleting server $this ..."
    }
    # Delete handlers before deleting event-generating members
    foreach handler $myEventHandlers {
        $handler delete
    }
    if {[info exists qhandler]} {
        $qhandler delete
    }
    if {[info exists thandler]} {
        $thandler delete
    }
    if {$imadelog} {
        $log delete
    }
    if {$imadeprotocol} {
        $protocol delete
    }
  }

}
